@@ -1,14 +1,12 @@
-# Copyright (c) 2000-2002 Graham Barr <gbarr@pobox.com>. All rights reserved.
+# Copyright (c) 2000-2005 Graham Barr <gbarr@pobox.com>. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
 package Convert::ASN1;
 
-# $Id: _decode.pm,v 1.15 2002/03/25 09:06:16 gbarr Exp $
-
 BEGIN {
   local $SIG{__DIE__};
-  eval { require bytes } and 'bytes'->import
+  eval { require bytes and 'bytes'->import };
 }
 
 # These are the subs that do the decode, they are called with
@@ -33,6 +31,7 @@ my @decode = (
   undef, # ANY
   undef, # CHOICE
   \&_dec_object_id,
+  \&_dec_bcd,
 );
 
 my @ctr;
@@ -68,7 +67,7 @@ sub _decode {
 	      # should be getting undef. So if it does not get undef
 	      # it knows it has no variable
 	      ($seqof ? $seqof->[$idx++] : defined($var) ? $stash->{$var} : ref($stash) eq 'SCALAR' ? $$stash : 1),
-	      $buf,$npos,$len, $indef ? $larr : []
+	      $buf,$npos,$len, $larr
 	    );
 
 	    $pos = $npos+$len+$indef;
@@ -87,7 +86,7 @@ sub _decode {
 	      $npos,
 	      $npos+$len,
 	      (\my @ctrlist),
-	      $indef ? $larr : [],
+	      $larr,
 	      $buf,
 	    );
 
@@ -101,11 +100,10 @@ sub _decode {
 	  }
 
 	  if ($seqof || defined $op->[cOPT]) {
-	    unshift @$larr, $len if $indef;
 	    next OP;
 	  }
 
-	  die "decode error " . unpack("H*",$tag) ."<=>" . unpack("H*",$op->[cTAG]);
+	  die "decode error " . unpack("H*",$tag) ."<=>" . unpack("H*",$op->[cTAG]), " ",$pos," ",$op->[cTYPE]," ",$op->[cVAR]||'';
         }
       }
       else { # opTag length is zero, so it must be an ANY or CHOICE
@@ -122,8 +120,13 @@ sub _decode {
 
 	    $len += $npos-$pos;
 
+             if ($op->[cDEFINE]) {
+                $handler = $optn->{oidtable} && $optn->{oidtable}{$stash->{$op->[cDEFINE]}};
+                $handler ||= $optn->{handlers}{$op->[cVAR]}{$stash->{$op->[cDEFINE]}};
+             }
+
 	    ($seqof ? $seqof->[$idx++] : ref($stash) eq 'SCALAR' ? $$stash : $stash->{$var})
-	      = substr($buf,$pos,$len);
+	      = $handler ? $handler->decode(substr($buf,$pos,$len)) : substr($buf,$pos,$len);
 
 	    $pos += $len + $indef;
 
@@ -153,8 +156,8 @@ sub _decode {
 		  $optn,
 		  $cop,
 		  $nstash,
-		  $nstash->{$cop->[cVAR]},
-		  $buf,$npos,$len,$indef ? $larr : []
+		  ($cop->[cVAR] ? $nstash->{$cop->[cVAR]} : undef),
+		  $buf,$npos,$len,$larr,
 		);
 
 		$pos = $npos+$len+$indef;
@@ -172,7 +175,7 @@ sub _decode {
 		    $pos,
 		    $npos+$len+$indef,
 		    undef,
-		    $indef ? $larr : [],
+		    $larr,
 		    $buf,
 		  );
 
@@ -210,7 +213,7 @@ sub _decode {
 		  $npos,
 		  $npos+$len,
 		  (\my @ctrlist),
-		  $indef ? $larr : [],
+		  $larr,
 		  $buf,
 		);
 
@@ -278,7 +281,7 @@ sub _dec_null {
 # 0      1    2       3     4     5     6
 # $optn, $op, $stash, $var, $buf, $pos, $len
 
-  $_[3] = 1;
+  $_[3] = exists($_[0]->{decode_null}) ? $_[0]->{decode_null} : 1;
   1;
 }
 
@@ -288,8 +291,17 @@ sub _dec_object_id {
 # $optn, $op, $stash, $var, $buf, $pos, $len
 
   my @data = unpack("w*",substr($_[4],$_[5],$_[6]));
-  splice(@data,0,1,int($data[0]/40),$data[0] % 40)
-    if $_[1]->[cTYPE] == opOBJID and $data[0];
+  if ($_[1]->[cTYPE] == opOBJID and @data > 1) {
+    if ($data[0] < 40) {
+      splice(@data, 0, 1, 0, $data[0]);
+    }
+    elsif ($data[0] < 80) {
+      splice(@data, 0, 1, 1, $data[0] - 40);
+    }
+    else {
+      splice(@data, 0, 1, 2, $data[0] - 80);
+    }
+  }
   $_[3] = join(".", @data);
   1;
 }
@@ -402,7 +414,7 @@ SET_OP:
 	    # should be getting undef. So if it does not get undef
 	    # it knows it has no variable
 	    (defined($var) ? $stash->{$var} : 1),
-	    $_[4],$npos,$len,$indef ? $larr : []
+	    $_[4],$npos,$len,$larr,
 	  );
 	  $done = $idx;
 	  last SET_OP;
@@ -417,7 +429,7 @@ SET_OP:
 	    $npos,
 	    $npos+$len,
 	    (\my @ctrlist),
-	    $indef ? $larr : [],
+	    $larr,
 	    $_[4],
 	  );
 
@@ -441,7 +453,7 @@ SET_OP:
 	      $cop,
 	      $nstash,
 	      $nstash->{$cop->[cVAR]},
-	      $_[4],$npos,$len,$indef ? $larr : []
+	      $_[4],$npos,$len,$larr,
 	    );
 	    $done = $idx;
 	    last SET_OP;
@@ -458,7 +470,7 @@ SET_OP:
 	      $npos,
 	      $npos+$len,
 	      (\my @ctrlist),
-	      $indef ? $larr : [],
+	      $larr,
 	      $_[4],
 	    );
 
@@ -502,7 +514,7 @@ sub _dec_time {
 
   my $mode = $_dec_time_opt{$_[0]->{'decode_time'} || ''} || 0;
 
-  if ($mode == 2) {
+  if ($mode == 2 or $_[6] == 0) {
     $_[3] = substr($_[4],$_[5],$_[6]);
     return;
   }
@@ -536,12 +548,20 @@ sub _dec_utf8 {
 # $optn, $op, $stash, $var, $buf, $pos, $len
 
   BEGIN {
-    local $SIG{__DIE__};
-    eval { require bytes } and 'bytes'->unimport;
-    eval { require utf8  } and 'utf8'->import;
+    unless (CHECK_UTF8) {
+      local $SIG{__DIE__};
+      eval { require bytes } and 'bytes'->unimport;
+      eval { require utf8  } and 'utf8'->import;
+    }
+  }
+
+  if (CHECK_UTF8) {
+    $_[3] = Encode::decode('utf8', substr($_[4],$_[5],$_[6]));
+  }
+  else {
+    $_[3] = (substr($_[4],$_[5],$_[6]) =~ /(.*)/s)[0];
   }
 
-  $_[3] = (substr($_[4],$_[5],$_[6]) =~ /(.*)/s)[0];
   1;
 }
 
@@ -574,11 +594,11 @@ sub _decode_tl {
       ($len,$pos) = (unpack("N", "\0" x (4 - $len) . substr($_[0],$pos,$len)), $pos+$len);
     }
     else {
-      unless (@$larr) {
+      unless (exists $larr->{$pos}) {
         _scan_indef($_[0],$pos,$end,$larr) or return;
       }
       $indef = 2;
-      $len = shift @$larr;
+      $len = $larr->{$pos};
     }
   }
 
@@ -592,8 +612,7 @@ sub _decode_tl {
 
 sub _scan_indef {
   my($pos,$end,$larr) = @_[1,2,3];
-  @$larr = ( $pos );
-  my @depth = ( \$larr->[0] );
+  my @depth = ( $pos );
 
   while(@depth) {
     return if $pos+2 > $end;
@@ -602,7 +621,7 @@ sub _scan_indef {
       my $end = $pos;
       my $stref = shift @depth;
       # replace pos with length = end - pos
-      $$stref = $end - $$stref;
+      $larr->{$stref} = $end - $stref;
       $pos += 2;
       next;
     }
@@ -628,8 +647,7 @@ sub _scan_indef {
       }
       else {
         # reserve another list element
-        push @$larr, $pos; 
-        unshift @depth, \$larr->[-1];
+        unshift @depth, $pos;
       }
     }
     else {
@@ -646,5 +664,12 @@ sub _ctr_bitstring {
   [ join('', map { $_->[0] } @_), $_[-1]->[1] ]
 }
 
+sub _dec_bcd {
+# 0      1    2       3     4     5     6
+# $optn, $op, $stash, $var, $buf, $pos, $len
+
+  ($_[3] = unpack("H*", substr($_[4],$_[5],$_[6]))) =~ s/[fF]$//;
+  1;
+}
 1;
 
